@@ -630,8 +630,8 @@ class ChunkedArray(awkward.array.base.AwkwardArray):
         out.knowchunksizes()
         return out
 
-    def pad(self, length, maskedwhen=True, clip=False):
-        return self.copy(chunks=[self._util_pad(x, length, maskedwhen, clip) for x in self._chunks], chunksizes=self.chunksizes)
+    def pad(self, length, maskedwhen=True, clip=False, axis=0):
+        return self.copy(chunks=[self._util_pad(x, length, maskedwhen, clip, axis) for x in self._chunks], chunksizes=self.chunksizes)
 
     def regular(self):
         self._valid()
