@@ -2,6 +2,7 @@
 
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE
 
+import codecs
 import json
 
 import numpy
@@ -81,7 +82,7 @@ def schema2type(schema):
                 return awkward.type.OptionType(out)
             else:
                 return out
-            
+
         elif isinstance(tpe, pyarrow.lib.DataType):
             if nullable:
                 return awkward.type.OptionType(tpe.to_pandas_dtype())
@@ -132,7 +133,7 @@ def toarrow(obj):
         elif isinstance(obj, awkward.array.masked.IndexedMaskedArray):
             thismask = obj.boolmask(maskedwhen=True)
             if mask is not None:
-                thismask = mask & thismask
+                thismask = mask | thismask
             if len(obj.content) == 0:
                 content = obj.numpy.empty(len(obj.mask), dtype=obj.DEFAULTTYPE)
             else:
@@ -142,24 +143,30 @@ def toarrow(obj):
         elif isinstance(obj, awkward.array.masked.MaskedArray):   # includes BitMaskedArray
             thismask = obj.boolmask(maskedwhen=True)
             if mask is not None:
-                thismask = mask & thismask
+                thismask = mask | thismask
             return recurse(obj.content, thismask)
 
+        elif isinstance(obj, awkward.array.objects.StringArray):
+            # # FIXME: BinaryArray.from_buffers is not implemented in pyarrow yet.
+            # if obj.encoding is None:
+            #     convert = lambda length, offsets, content: pyarrow.BinaryArray.from_buffers(pyarrow.binary(), length, [None, offsets, content])
+            # elif codecs.lookup(obj.encoding) is codecs.lookup("utf-8"):
+            #     convert = lambda length, offsets, content: pyarrow.StringArray.from_buffers(length, offsets, content)
+            # else:
+            #     raise ValueError("only encoding=None or encoding='utf-8' can be converted to Arrow")
+            convert = lambda length, offsets, content: pyarrow.StringArray.from_buffers(length, offsets, content)
+
+            obj = obj.compact()
+            offsets = obj.offsets
+            if offsets.dtype != numpy.dtype(numpy.int32):
+                offsets = offsets.astype(numpy.int32)
+
+            return convert(len(offsets) - 1, pyarrow.py_buffer(offsets), pyarrow.py_buffer(obj.content))
+
         elif isinstance(obj, awkward.array.objects.ObjectArray):
             # throw away Python object interpretation, which Arrow can't handle while being multilingual
             return recurse(obj.content, mask)
 
-        elif isinstance(obj, awkward.array.objects.StringArray):
-            # obj = obj.compact()
-            raise NotImplementedError("I don't know how to make an Arrow StringArray")
-
-            # I don't understand this
-            # pyarrow.StringArray.from_buffers(2, pyarrow.py_buffer(numpy.array([0, 5, 10])), pyarrow.py_buffer(b"helloHELLO"), offset=0)
-            # returns ["", "hello"]
-            # ???
-
-            # Also, be sure to check for the difference between strings and bytes!
-
         elif isinstance(obj, awkward.array.table.Table):
             return pyarrow.StructArray.from_arrays([recurse(x, mask) for x in obj.contents.values()], list(obj.contents))
 
@@ -218,20 +225,26 @@ def fromarrow(obj, awkwardlib=None):
     ARROW_TAGTYPE = awkwardlib.numpy.uint8
     ARROW_CHARTYPE = awkwardlib.numpy.uint8
 
-    def popbuffers(tpe, buffers, length):
+    def popbuffers(array, tpe, buffers, length):
         if isinstance(tpe, pyarrow.lib.DictionaryType):
-            index = popbuffers(tpe.index_type, buffers, length)
-            content = fromarrow(tpe.dictionary)
+            index = popbuffers(None if array is None else array.indices, tpe.index_type, buffers, length)
+            if hasattr(tpe, "dictionary"):
+                content = fromarrow(tpe.dictionary)
+            elif array is not None:
+                content = fromarrow(array.dictionary)
+            else:
+                raise NotImplementedError("no way to access Arrow dictionary inside of UnionArray")
             if isinstance(index, awkwardlib.BitMaskedArray):
                 return awkwardlib.BitMaskedArray(index.mask, awkwardlib.IndexedArray(index.content, content), maskedwhen=index.maskedwhen, lsborder=index.lsborder)
             else:
                 return awkwardlib.IndexedArray(index, content)
 
         elif isinstance(tpe, pyarrow.lib.StructType):
+            assert getattr(tpe, "num_buffers", 1) == 1
             mask = buffers.pop(0)
             pairs = []
             for i in range(tpe.num_children):
-                pairs.append((tpe[i].name, popbuffers(tpe[i].type, buffers, length)))
+                pairs.append((tpe[i].name, popbuffers(None if array is None else array.field(tpe[i].name), tpe[i].type, buffers, length)))
             out = awkwardlib.Table.frompairs(pairs, 0)   # FIXME: better rowstart
             if mask is not None:
                 mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
@@ -240,9 +253,10 @@ def fromarrow(obj, awkwardlib=None):
                 return out
 
         elif isinstance(tpe, pyarrow.lib.ListType):
+            assert getattr(tpe, "num_buffers", 2) == 2
             mask = buffers.pop(0)
             offsets = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length + 1]
-            content = popbuffers(tpe.value_type, buffers, offsets[-1])
+            content = popbuffers(None if array is None else array.flatten(), tpe.value_type, buffers, offsets[-1])
             out = awkwardlib.JaggedArray.fromoffsets(offsets, content)
             if mask is not None:
                 mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
@@ -251,6 +265,7 @@ def fromarrow(obj, awkwardlib=None):
                 return out
 
         elif isinstance(tpe, pyarrow.lib.UnionType) and tpe.mode == "sparse":
+            assert getattr(tpe, "num_buffers", 3) == 3
             mask = buffers.pop(0)
             tags = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_TAGTYPE)[:length]
             assert buffers.pop(0) is None
@@ -261,7 +276,7 @@ def fromarrow(obj, awkwardlib=None):
                     sublength = index[tags == i][-1] + 1
                 except IndexError:
                     sublength = 0
-                contents.append(popbuffers(tpe[i].type, buffers, sublength))
+                contents.append(popbuffers(None, tpe[i].type, buffers, sublength))
             for i in range(len(contents)):
                 these = index[tags == i]
                 if len(these) == 0:
@@ -276,6 +291,7 @@ def fromarrow(obj, awkwardlib=None):
                 return out
 
         elif isinstance(tpe, pyarrow.lib.UnionType) and tpe.mode == "dense":
+            assert getattr(tpe, "num_buffers", 3) == 3
             mask = buffers.pop(0)
             tags = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_TAGTYPE)[:length]
             index = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length]
@@ -285,7 +301,7 @@ def fromarrow(obj, awkwardlib=None):
                     sublength = index[tags == i].max() + 1
                 except ValueError:
                     sublength = 0
-                contents.append(popbuffers(tpe[i].type, buffers, sublength))
+                contents.append(popbuffers(None, tpe[i].type, buffers, sublength))
             for i in range(len(contents)):
                 these = index[tags == i]
                 if len(these) == 0:
@@ -300,6 +316,7 @@ def fromarrow(obj, awkwardlib=None):
                 return out
 
         elif tpe == pyarrow.string():
+            assert getattr(tpe, "num_buffers", 3) == 3
             mask = buffers.pop(0)
             offsets = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length + 1]
             content = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_CHARTYPE)[:offsets[-1]]
@@ -311,6 +328,7 @@ def fromarrow(obj, awkwardlib=None):
                 return out
 
         elif tpe == pyarrow.binary():
+            assert getattr(tpe, "num_buffers", 3) == 3
             mask = buffers.pop(0)
             offsets = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length + 1]
             content = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_CHARTYPE)[:offsets[-1]]
@@ -322,6 +340,7 @@ def fromarrow(obj, awkwardlib=None):
                 return out
 
         elif tpe == pyarrow.bool_():
+            assert getattr(tpe, "num_buffers", 2) == 2
             mask = buffers.pop(0)
             out = awkwardlib.numpy.unpackbits(awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_CHARTYPE)).view(awkwardlib.MaskedArray.BOOLTYPE)
             out = out.reshape(-1, 8)[:,::-1].reshape(-1)[:length]    # lsborder=True
@@ -332,6 +351,7 @@ def fromarrow(obj, awkwardlib=None):
                 return out
 
         elif isinstance(tpe, pyarrow.lib.DataType):
+            assert getattr(tpe, "num_buffers", 2) == 2
             mask = buffers.pop(0)
             out = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=tpe.to_pandas_dtype())[:length]
             if mask is not None:
@@ -345,7 +365,7 @@ def fromarrow(obj, awkwardlib=None):
 
     if isinstance(obj, pyarrow.lib.Array):
         buffers = obj.buffers()
-        out = popbuffers(obj.type, buffers, len(obj))
+        out = popbuffers(obj, obj.type, buffers, len(obj))
         assert len(buffers) == 0
         return out
 
@@ -464,7 +484,7 @@ class _ParquetFile(object):
         import pyarrow.parquet
         self.parquetfile = pyarrow.parquet.ParquetFile(self.file, metadata=self.metadata, common_metadata=self.common_metadata)
         self.type = schema2type(self.parquetfile.schema.to_arrow_schema())
-        
+
     def __getstate__(self):
         return {"file": self.file, "metadata": self.metadata, "common_metadata": self.common_metadata}
 
